#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>

#include <glib.h>

#include <utillib.h>

#include "debug.h"
#include "download.h"
#include "http.h"
#include "ftp.h"

#include "downmand-marshal.h"

#include "download-thread.h"

#define DEFAULT_FTP_PORT 21
#define DEFAULT_HTTP_PORT 80


GType           download_get_type           ();
static void     download_class_init         (DownloadClass *klass);
static void     download_init               (Download *download);
static void     download_finalize           (GObject* object);

void            download_thread_exit        (Download *download,
                                             DownloadThread *thread);

enum {
    DOWNLOAD_UPDATE,
    DOWNLOAD_EXIT,
    LAST_SIGNAL
};

static gint download_signals[LAST_SIGNAL] = { 0 };

static gpointer parent_class = NULL;


gpointer download_thread(gpointer data);

void    download_cleanup        (void *arg);
gint    resolve                 (Download *download, DownloadThread *thread);
gint    create_socket           (Download *download, DownloadThread *thread);
gint    connect_to_host         (Download *download, DownloadThread *thread);

void    download_mutex_lock     (Download *download);
void    download_mutex_unlock   (Download *download);


GType download_get_type ()
{
    static GType download_type = 0;

    if (!download_type)
    {
        static const GTypeInfo download_info =
        {
            sizeof (DownloadClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) download_class_init,
            NULL,           /* class_finalize */
            NULL,           /* class_data */
            sizeof (Download),
            0,              /* n_preallocs */
            (GInstanceInitFunc) download_init
        };

        download_type = g_type_register_static (G_TYPE_OBJECT,
                                               "Download",
                                               &download_info,
                                               0);
    }

    return download_type;
}

static void download_class_init (DownloadClass *klass)
{
    GObjectClass *object_class;

    object_class = (GObjectClass*) klass;

    parent_class = g_type_class_peek_parent (klass);

    download_signals[DOWNLOAD_UPDATE] = g_signal_new ("download_update",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (DownloadClass, download_update_signal),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    download_signals[DOWNLOAD_EXIT] = g_signal_new ("download_exit",
            G_OBJECT_CLASS_TYPE (klass),
            G_SIGNAL_RUN_FIRST,
            G_STRUCT_OFFSET (DownloadClass, download_exit_signal),
            NULL, NULL,
            downmand_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);

    klass->download_update_signal = NULL;
    klass->download_exit_signal = NULL;

    object_class->finalize  = download_finalize;
}

static void download_init (Download *download)
{
	download->id = 0;

	download->threads = list_new();
	download->addr = 0;

	download->bytes_done = 0;
	download->bytes_total = 0;
	download->speed = 0;
    download->speedlimit = 0;
    download->real_speedlimit = 0;
	download->target_ready = 0;
	download->queue = DOWNLOAD_QUEUE_NONE;
	download->status = STATUS_QUEUED;
	download->exit_cause = DOWNLOAD_EXIT_NONE;

    // init mutex
    g_static_mutex_init (&download->download_mutex);
}

static void
download_finalize (GObject* object)
{
    Download *download = DOWNLOAD (object);

    g_print ("download_finalize\n");
    xmlFreeURI (download->url);
    g_free (download->url_string);

    if (G_OBJECT_CLASS (parent_class)->finalize)
        (* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

Download *download_new(gchar *url)
{
    Download *download = DOWNLOAD (g_object_new (download_get_type (), NULL));

    if (url != NULL)
    {
        download_url_set (download, url);
    }

    download->linelog = linelog_new ();

	return download;
}

void
download_url_set (Download *download, gchar *url)
{
    download->url_string = g_strdup (url);

	download->url = xmlParseURI(url);

    if (download->url == NULL)
    {
        g_print ("ERROR parsing url\n");
        return;
    }

/*
    g_print ("=====================================================\n");
    g_print ("scheme = %s\n", download->url->scheme);
    g_print ("opaque = %s\n", download->url->opaque);
    g_print ("authority = %s\n", download->url->authority);
    g_print ("server = %s\n", download->url->server);
    g_print ("user = %s\n", download->url->user);
    g_print ("port = %i\n", download->url->port);
    g_print ("path = %s\n", download->url->path);
    g_print ("query = %s\n", download->url->query);
    g_print ("fragment = %s\n", download->url->fragment);
    g_print ("cleanup = %i\n", download->url->cleanup);
    g_print ("=====================================================\n");
*/
}

void
download_path_set (Download *download, gchar *path)
{
    if (download->local_path)
        g_free (download->local_path);

    download->local_path = g_strdup (path);
}

void download_start(Download *download, GError **error)
{
    GThread *thread;
    GError *err = NULL;

    if (!download->url)
    {
		g_set_error (error, DM_DOWNMAND_ERROR, 1,
			"The URL is invalid");
        return;
    }

    if (!download->url->scheme)
    {
		g_set_error (error, DM_DOWNMAND_ERROR, 1,
			"The URL scheme is invalid");
        return;
    }

    if (!(strncasecmp(download->url->scheme, "HTTP", 4) == 0) &&
        !(strncasecmp(download->url->scheme, "FTP", 3) == 0))
    {
		g_set_error (error, DM_DOWNMAND_ERROR, 1,
			"Protocol not supported");
        return;
    }

    if (!download->url->server)
    {
		g_set_error (error, DM_DOWNMAND_ERROR, 1,
			"There is no server to connect to");
        return;
    }

    download_mutex_lock(download);

    thread = g_thread_create (download_thread, download,
                              FALSE, &err);

    if (thread == NULL)
    {
        g_debug("Thread creation failed: %s\n", err->message);
    }

    download_mutex_unlock(download);
}

gpointer download_thread(gpointer data)
{
    Download *download = DOWNLOAD(data);
    DownloadThread *thread = download_thread_new();
	xmlURIPtr url = download->url;
	GString *buffer = g_string_new (NULL);
    GString *logger;
    GError *error = NULL;
	int select_fds, count = -1;
	struct timeval timeout, oldtime = {0,0}, newtime;
	char inbuffer[4100];
	fd_set readfds;
	fd_set waitfd;
	unsigned long old_size = 0, new_size = 0, speed = 0;
	ulong file_size = 0, file_offset = 0;
	int speeds[5];
	struct stat *file_stat = malloc(sizeof(struct stat));
    gint speedlimit_left;

	download->status = STATUS_INIT;
	download->exit_cause = DOWNLOAD_EXIT_NONE;

    list_add(download->threads, list_node_new(thread));

	thread->sock_fd = -1;
	thread->data_fd = -1;
	thread->file_fd = -1;

    linelog_disklog (download->linelog, TRUE, &error);
    error = NULL;

    download_log_add (download, "* Starting download\n");

    if (download->target_ready == 0)
    {
        // Check if we can write to the target file
    }

    if ((!url->path) || (url->path[strlen (url->path) - 1] == '/'))
    {
        g_string_printf(buffer, "%s/%s",
                        download->local_path, "index.html");
    }
    else
    {
        g_string_printf(buffer, "%s/%s",
                        download->local_path,
                        strrchr (url->path,'/'));
    }

    if (stat(buffer->str, file_stat) < 0)
    {
        thread->file_fd = open(buffer->str, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR);
        if (thread->file_fd < 0)
        {
            download->status = ERROR_OPENING_FILE;
        }
        else
        {
            old_size = new_size = 0;
        }
    }
    else
    {
        thread->file_fd = open(buffer->str, O_WRONLY, S_IRUSR | S_IWUSR);
        if (thread->file_fd < 0)
        {
            download->status = ERROR_OPENING_FILE;
        }
        else
        {
            file_offset = file_stat->st_size;
            file_offset = lseek (thread->file_fd, 0, SEEK_END);
            download->bytes_done = old_size = new_size = file_offset;
        }
    }

    g_string_free(buffer, TRUE);

    if (thread->file_fd < 0)
    {
        download_thread_exit(download, thread);
    }

    if (thread->stop)
        download_thread_exit(download, thread);

    /* Destination file is ok and opened, go on */
    download->status = STATUS_STARTING;

    /* create socket, resolve name and connect to host */
    if (create_socket(download, thread))
    {
        download_thread_exit(download, thread);
    }

    if (thread->stop)
        download_thread_exit(download, thread);

    logger = g_string_new (NULL);
    g_string_printf (logger, "* Resolving %s\n", url->server);
    linelog_add (download->linelog, logger, &error);

    download->status = STATUS_RESOLVING;
    if (resolve(download, thread))
    {
        download_thread_exit(download, thread);
    }

    if (thread->stop)
        download_thread_exit(download, thread);

    logger = g_string_new (NULL);
    g_string_printf (logger, "* Connecting to %s\n", url->server);
    linelog_add (download->linelog, logger, &error);

    download->status = STATUS_CONNECTING;
    if (connect_to_host(download, thread))
    {
        download_thread_exit(download, thread);
    }

    if (thread->stop)
        download_thread_exit(download, thread);

	if (strncasecmp(url->scheme, "HTTP", 4) == 0)
	{
        download->status = STATUS_WAITING_DATA;
	}
    else if (strncasecmp(url->scheme, "FTP", 3) == 0)
    {
        download->status = STATUS_LOGIN;
    }

	// log into the server and get the file
	if (strncasecmp(url->scheme, "HTTP", 4) == 0)
	{
		GIOChannel *data_channel = g_io_channel_unix_new (thread->sock_fd);

		http_get (thread->sock_fd, url, file_offset, download->linelog);
		thread->data_fd = thread->sock_fd;
		thread->sock_fd = -1;

        http_get_head (data_channel, &file_size, download->linelog);
        if (download->bytes_total > 0)
            file_size = download->bytes_total;
        else
            download->bytes_total = file_size;

		/* write the buffered data */
        write (thread->file_fd, data_channel->read_buf->str,
               data_channel->read_buf->len);
		new_size += data_channel->read_buf->len;
		download->bytes_done += data_channel->read_buf->len;
	}
    else if (strncasecmp(url->scheme, "FTP", 3) == 0)
    {
        if (ftp_login (thread->sock_fd, url, download->linelog, &error))
        {
//FIXME: some way to tell the user the cause of the error.            
            download->status = ERROR_LOGIN;
            download_thread_exit(download, thread);
        }

        if (download->bytes_total == 0)
        {
            if (ftp_get_file_size (thread->sock_fd, url, &file_size,
                download->linelog))
            {
                download->status = ERROR_GETTING_HEAD;
                download_thread_exit(download, thread);
            }
            else
            {
                if (download->bytes_total > 0)
                    file_size = download->bytes_total;
                else
                    download->bytes_total = file_size;
            }
        }

        ftp_get (thread->sock_fd, url, &thread->data_fd, file_offset,
                 download->linelog);
    }

    if (thread->stop)
        download_thread_exit(download, thread);

    /* Start getting data */
	download->status = STATUS_TRANSFERING;

    logger = g_string_new (NULL);
    g_string_printf (logger, "* Getting data from %li to %li\n", file_offset,
                     download->bytes_total);
    linelog_add (download->linelog, logger, &error);

    /* Check if the size is known or unknown */
    if (file_size > 0)
    {
        if (file_offset >= download->bytes_total)
        {
            download->exit_cause = DOWNLOAD_EXIT_STOPPED;
//FIXME: better error handling
            download_thread_exit(download, thread);
        }
    }

    speedlimit_left = download->real_speedlimit;

	while ((count > 0) || (count == -1))
	{
		FD_ZERO (&readfds);
		FD_SET (thread->data_fd, &readfds);

		timeout.tv_sec = 1;
		timeout.tv_usec = 0;

		/* wait for some input */
		select_fds = select (FD_SETSIZE, &readfds, NULL, NULL, &timeout);

		/* look for some data */
		if (select_fds > 0)
		{
            if (download->real_speedlimit)
            {
                while (speedlimit_left <= 0)
                {
                    FD_ZERO (&waitfd);
                    timeout.tv_sec = 0;
                    timeout.tv_usec = 1000;
                    select_fds = select (FD_SETSIZE, &waitfd, NULL, NULL, &timeout);
                    gettimeofday(&newtime, NULL);
                    if (newtime.tv_sec > oldtime.tv_sec)
                        speedlimit_left += download->real_speedlimit;
                }
                count = read (thread->data_fd, inbuffer, 1024);
                inbuffer[count] = '\0';
                speedlimit_left -= count;
            }
            else
            {
                count = read (thread->data_fd, inbuffer, 4096);
                inbuffer[count] = '\0';
            }
		}
		else
		{
			/* no data, is there a problem? */
			if (select_fds == -1)
			{
				download->status = ERROR_GETTING_DATA;
                download_thread_exit(download, thread);
			}
			count = -1;
		}

		/* write the read data to the file */
		if (count > 0)
		{
			write(thread->file_fd, inbuffer, count);
			new_size += count;
			download->bytes_done += count;
		}
	
		/* some stadistics */
		gettimeofday(&newtime, NULL);
		if (newtime.tv_sec > oldtime.tv_sec)
		{
			int i;
			oldtime = newtime;
			speeds[newtime.tv_sec % 5] = (new_size - old_size);
			speed = 0;
			for (i = 0; i < 5; i++)
			{
				speed += speeds[i];
			}
			speed /= 5;
			download->speed = speed;
/*
			if (file_size)
				printf("\r%.2f Kb/s - %li of %li - %.2f %% (%i)       ", (float)((float)speed / (float)1024),
						new_size, file_size, ((double)new_size/file_size)*100, count);
			else
				printf("\r%.2f Kb/s - %li (%i)        ", (float)((float)speed / (float)1024), new_size, count);
			fflush(stdout);
*/
			old_size = new_size;
		}

        /* Check if we need to exit */
        if (thread->stop)
            download_thread_exit(download, thread);
	}

    logger = g_string_new (NULL);
    g_string_printf (logger, "* Download end\n");
    linelog_add (download->linelog, logger, &error);

    /* Check if we have completed or the size is unknown */
    if ((new_size == file_size) || (file_size == 0))
    {
        download->exit_cause = DOWNLOAD_EXIT_DONE;
        download->bytes_total = new_size;
    }
    else
    {
        download->status = ERROR_NOT_COMPLETE;
    }
    
    download_thread_exit(download, thread);

    return NULL;
}

void download_thread_exit(Download *download, DownloadThread *thread)
{
    DownloadThread *list_thread;
    Node *element;
    GError *error;

//FIXME: per thread exit_cause
    switch (download->exit_cause)
    {
        case DOWNLOAD_EXIT_STOPPED:
            download->status = STATUS_STOPPED;
            break;
        case DOWNLOAD_EXIT_QUEUED:
            download->status = STATUS_QUEUED;
            break;
        case DOWNLOAD_EXIT_DONE:
            download->status = STATUS_DONE;
            break;
        case DOWNLOAD_EXIT_SHUTDOWN:
            download->status = STATUS_STOPPED;
            break;

        default:
            switch(download->status)
            {
                case ERROR_OPENING_FILE:
                    sleep(10);
                    break;

                case ERROR_CREATING_SOCKET:
                    sleep(10);
                    break;

                case ERROR_RESOLVING:
                    sleep(10);
                    break;

                case ERROR_CONNECTING:
                    sleep(5);
                    break;

                case ERROR_LOGIN:
                    sleep (20);
                    break;

                case ERROR_GETTING_HEAD:
                    sleep (10);
                    break;

                case ERROR_GETTING_DATA:
                    sleep (10);
                    break;

                case ERROR_NOT_COMPLETE:
                    sleep (10);
                    break;

                default:
                    sleep (10);
                    break;
            }
            break;
    }

    if (thread->sock_fd > -1)
        close(thread->sock_fd);

    if (thread->data_fd > -1)
        close(thread->data_fd);

    if (thread->file_fd > -1)
        close(thread->file_fd);

    download_mutex_lock(download);
    element = list_begin(download->threads);
    while (element != 0)
    {
        list_thread = (DownloadThread *)list_node_data(element);
        if (list_thread == thread)
        {
            list_remove (download->threads, element);
            break;
        }

        element = list_next(element);
    }
    download_mutex_unlock(download);

    download->speed = 0;

    g_signal_emit (G_OBJECT (download),
               download_signals[DOWNLOAD_EXIT],
               0);

    linelog_disklog (download->linelog, FALSE, &error);

    g_thread_exit(NULL);
}

gint create_socket(Download *download, DownloadThread *thread)
{
	thread->sock_fd = socket(PF_INET, SOCK_STREAM, 0);

	if (thread->sock_fd < 0)
	{
		if ((debug & DM_DEBUG) && (debug & DM_DEBUG_SCHEDULER))
		{
			g_debug("DOWNLOAD: Can't create socket!\n");
		}
		download->status = ERROR_CREATING_SOCKET;
		return 1;
	}

    return 0;
}

gint resolve(Download *download, DownloadThread *thread)
{
	xmlURIPtr url = download->url;
	struct sockaddr_in addr;
	struct hostent hostbuf, *host;
	size_t bufferlen;
	char *buffer;
	int result, error;
	struct timeval newtime;

	gettimeofday(&newtime, NULL);

	if (thread->redirect == 0)
	{
		if (download->addr)
		{
			// if we resolved less than 5 minutes ago, use cached value.
			if ((newtime.tv_sec - download->resolve_time.tv_sec) < 60*5)
				return 0;
		}
	}
	else
	{
		if (thread->addr)
		{
			// if we resolved less than 5 minutes ago, use cached value.
			if ((newtime.tv_sec - thread->resolve_time.tv_sec) < 60*5)
				return 0;
		}
	}

	// resolve
	bufferlen = 1024;
	buffer = malloc(bufferlen);

//FIXME: use redirected url if needed
	while ((result = gethostbyname_r (url->server, &hostbuf, buffer, bufferlen,
									&host, &error)) == ERANGE)
	{
		bufferlen *= 2;
		buffer = realloc(buffer, bufferlen);
	}

	if (result || host == NULL)
	{
		if ((debug & DM_DEBUG) && (debug & DM_DEBUG_SCHEDULER))
		{
			g_debug("DOWNLOAD: %s\n", hstrerror(error));
		}
		close(thread->sock_fd);
		thread->sock_fd = -1;
		download->status = ERROR_RESOLVING;
		free(buffer);
		return 1;
	}

	addr.sin_family = AF_INET;
	if (url->port > 0)
		addr.sin_port = htons(url->port);
	else
	{
		if (strncasecmp(url->scheme, "HTTP", 4) == 0)
			addr.sin_port = htons(DEFAULT_HTTP_PORT);
		else if (strncasecmp(url->scheme, "FTP", 3) == 0)
			addr.sin_port = htons(DEFAULT_FTP_PORT);
	}

	memcpy(&addr.sin_addr, host->h_addr, host->h_length);

	free(buffer);

	if (thread->redirect == 0)
	{
		if (download->addr == 0)
			download->addr = malloc(sizeof(struct sockaddr_in));
//FIXME: lock download first
		download->resolve_time = newtime;
		*download->addr = addr;
	}
	else
	{
		if (thread->addr == 0)
			thread->addr = malloc(sizeof(struct sockaddr_in));

		thread->resolve_time = newtime;
		*thread->addr = addr;
	}

    return 0;
}

gint connect_to_host(Download *download, DownloadThread *thread)
{
	xmlURIPtr url = download->url;
	struct sockaddr_in *addr;

	if (thread->redirect == 0)
		addr = download->addr;
	else
		addr = thread->addr;

	// connect to the host
	if(connect(thread->sock_fd, (struct sockaddr *)addr, sizeof(*addr)) < 0)
	{
		int error = errno;

		if ((debug & DM_DEBUG) && (debug & DM_DEBUG_SCHEDULER))
		{
			g_debug("DOWNLOAD: Error connecting to server %s. Error was: %s\n",
                      url->server, strerror(error));
		}
		download->status = ERROR_CONNECTING;
		return 1;
	}

    return 0;
}

void download_stop(Download *download, DownloadStopAction action)
{
	DownloadThread *thread;
	Node *element;
	Node *element_next;

	switch(download->queue)
	{
		case DOWNLOAD_QUEUE_ACTIVE:
			switch(action)
			{
				case DOWNLOAD_STOP_NONE:
					download->exit_cause = DOWNLOAD_EXIT_STOPPED;
					break;
				case DOWNLOAD_STOP_QUEUE:
					download->exit_cause = DOWNLOAD_EXIT_QUEUED;
					break;
				case DOWNLOAD_STOP_DELETE:
					download->exit_cause = DOWNLOAD_EXIT_DELETE;
                    break;
                case DOWNLOAD_STOP_SHUTDOWN:
                    download->exit_cause = DOWNLOAD_EXIT_SHUTDOWN;
                    break;
				default:
					break;
			}
			download->status = STATUS_STOPPING;

//FIXME: create a threads_lock_mutex
			download_mutex_lock(download);
			element = list_begin(download->threads);

			while (element != 0)
			{
				element_next = list_next(element);

				thread = (DownloadThread *)list_node_data(element);
                thread->stop = 1;
//				list_remove(download->threads, element);

				element = element_next;
			}
			download_mutex_unlock(download);

			break;

		default:
			break;
	}

}

void
download_logdir_set (Download *download, gchar *path)
{
    GString *buffer = g_string_new (NULL);

	g_string_printf (buffer, "%s/%u.txt", path, download->id);
    linelog_setfile (download->linelog, buffer->str);

    g_string_free(buffer, TRUE);
}

void
download_log_add (Download *download, const gchar *format, ...)
{
    GString *logger = g_string_new (NULL);
    GError *error = NULL;
    va_list args;

    va_start (args, format);
    g_string_printf (logger, format, args);
    va_end (args);

    linelog_add (download->linelog, logger, &error);
//FIXME: log the error
}

gint download_status_get (Download *download)
{
    return download->status;
}

void download_status_set(Download *download, enum dm_download_status status)
{
	download->status = status;
}

guint
download_recursive_get (Download *download)
{
    return download->recursive;
}

void
download_recursive_set (Download *download, guint recursive)
{
    download->recursive = recursive;
}

void
download_default_path_set (Download *download, gboolean setting)
{
    download->default_path = setting;
}

void
download_speedlimit_set (Download *download, guint speed)
{
    download->speedlimit = speed;
}

guint
download_speedlimit_get (Download *download)
{
    return download->speedlimit;
}

void
download_real_speedlimit_set (Download *download, guint speed)
{
    download->real_speedlimit = speed;
}

guint
download_real_speedlimit_get (Download *download)
{
    return download->real_speedlimit;
}

DownloadExitCause download_exit_cause(Download *download)
{
	return download->exit_cause;
}

void
download_id_set (Download *download, guint id)
{
    download->id = id;
}

guint
download_id_get (Download *download)
{
    return download->id;
}

const char *download_url_string(Download *download)
{
	return download->url_string;
}

gulong download_bytes_done_get (Download *download)
{
	return download->bytes_done;
}

void download_bytes_done_set (Download *download, gulong bytes_done)
{
	download->bytes_done = bytes_done;
}

gulong download_bytes_total_get (Download *download)
{
	return download->bytes_total;
}

void download_bytes_total_set (Download *download, gulong bytes_total)
{
	download->bytes_total = bytes_total;
}

unsigned int download_speed(Download *download)
{
	return download->speed;
}

DownloadQueue download_queue_get (Download *download)
{
	return download->queue;
}

void download_queue_set (Download *download, DownloadQueue queue)
{
	download->queue = queue;
}

void download_mutex_lock(Download *download)
{
	g_static_mutex_lock(&download->download_mutex);
}

void download_mutex_unlock(Download *download)
{
	g_static_mutex_unlock(&download->download_mutex);
}
