#ifndef __PROJECT_H__
#define __PROJECT_H__

#include <glib-object.h>

#include "download.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#define PROJECT(obj)            G_TYPE_CHECK_INSTANCE_CAST (obj, project_get_type (), Project)
#define PROJECT_CLASS(klass)    G_TYPE_CHECK_CLASS_CAST (klass, project_get_type (), ProjectClass)
#define IS_PROJECT(obj)         G_TYPE_CHECK_INSTANCE_TYPE (obj, project_get_type ())


typedef struct _Project        Project;
typedef struct _ProjectClass   ProjectClass;

struct _Project
{
    GObject object;

    /*< private >*/

    GHashTable *hash;
    GHashTable *hash_active;

    List *batch;
    List *batch_active;

    GStaticMutex batch_mutex;
    GStaticMutex batch_active_mutex;

    /* Files of the project are saved in path */
    gchar *path;
    gchar *name;
    guint project_id;

    guint download_id;

    /* The pipe to comunicate the Project and the download */
    GIOChannel *listener_read;
    GIOChannel *listener_write;

    gint listen_tag;

    /* Counters */
    guint queued_downloads;

    /* Settings */
    guint max_downloads;
    guint auto_download;
    /* Downloaded files go to default_path */
    gchar *default_path;
    guint speedlimit;
};

struct _ProjectClass
{
    GObjectClass parent_class;

    /* emitted when a project changes the autodownload setting */
    void (* project_autodownload_changed)   (void);

    /* emitted when a project changes the maxdownloads setting */
    void (* project_maxdownloads_changed)   (void);

    /* emitted when a new download is added to the project */
    void (* download_new)       (Download *download);

    /* emitted when a download is stopped */
    void (* download_stop)      (Download *download);

    /* emitted when a download is finished */
    void (* download_done)      (Download *download);

    /* emitted when a download is removed from the project */
    void (* download_delete)    (Download *download);
};

GType           project_get_type        (void);
Project*        project_new             (guint project_id, GError **error);

void            project_name_set        (Project *project, const gchar *name);
const gchar*    project_name_get        (Project *project);
void            project_path_set        (Project *project, const gchar *path);
const gchar*    project_path_get        (Project *project);
void            project_default_path_set    (Project *project,
                                             const gchar *path);
const gchar*    project_default_path_get    (Project *project);
void            project_speedlimit_set  (Project *project, guint speedlimit);

GHashTable* project_hash            (Project *project);
List*       project_batch           (Project *project);
List*       project_batch_active    (Project *project);

guint       project_queued_downloads    (Project *project);

guint       project_size            (Project *project);
void        project_lock_batch      (Project *project);
void        project_unlock_batch    (Project *project);
void        project_lock_active     (Project *project);
void        project_unlock_active   (Project *project);

void    project_scheduler       (Project *project);
gint    project_command_download_new    (Project *project, gchar *command,
                                         GError **error);
gint    project_command_download_start  (Project *project, guint id,
                                         GError **error);
gint    project_command_download_stop   (Project *project, guint id,
                                         GError **error);
gint    project_command_download_queue  (Project *project, guint id,
                                         GError **error);
gint    project_command_download_delete (Project *project, guint id,
                                         GError **error);

gint    project_preferences_save    (Project *project);
gint    project_session_save        (Project *project);

gint    project_preferences_load    (Project *project);
gint    project_session_load        (Project *project);

guint           project_id          (Project *project);
const gchar*    project_name        (Project *project);
gint            project_auto        (Project *project);
void            project_auto_set    (Project *project, guint setting,
                                     GError **error);
guint           project_maxdownloads_get    (Project *project);
void            project_maxdownloads_set    (Project *project,
                                             guint maxdownloads,
                                             GError **error);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __PROJECT_H__ */
